/******************************************************************************
* NKMP_HESS.G
* computes the Hessian at the posterior mode
*
* upadted:  06/02/2005
******************************************************************************/

/********************************************************
* House cleaning and set defaults
********************************************************/
library pgraph, user;
cls;
outwidth 256;

pmsel   = "";

format /mat /on /mb1 /ros 16,8;

/** Global defaults **/
__output = 0;       /** 1 for verbose of CSMIN**/
__graph  = 1;       /** 1 for graph mode **/
_verbose = 2;       /** 1 for verbose **/
_fcmptol = 1E-10;

#include 1MSleadershipfmmod.src;
#include 2MSleadershipfmlh.src;
#include 3MSleadershipfmpara.src;
#include spec.g;
#include dsgesel2.src;
#include pathspec.g;
#include matop.src;

/** start log file **/
outstr = resupath $+ "\\MSleadershipfm_hess.log";
output file = ^outstr reset;

/********************************************************
*  Import prior specification
********************************************************/
priorfile = "MSleadershipfmused" $+ prioname $+ ".txt";
load path = ^priopath prior[] = ^priorfile;
prior = reshape(prior,rows(prior)/5,5);

{prior,para_names} = psel(prior,msel);

npara    = rows(prior);
pshape   = prior[.,1];
pmean    = prior[.,2];
pstdd    = prior[.,3];
pmask    = prior[.,4];
pfix     = prior[.,5];
pmaskinv = 1-pmask;
pshape   = pshape.*pmaskinv;

/********************************************************
* Import data series
********************************************************/
#include loaddata.g; 

nobs  = rows(YY);  /** number of observations **/
ny    = cols(YY);
/********************************************************
* Load Posterior Mode Estimates
********************************************************/
lpara = postpath $+ "\\MSleadershipfm_pm" $+ pmsel;
lretc = postpath $+ "\\MSleadershipfm_pmret" $+ pmsel;

/** check if pm results exist **/
if not(exist(lpara $+ ".dat"));
    "";
    "Please run posterior maximization program first.";
    "";
    end;
endif;

/** open **/
open fhpara = ^lpara for read;
para = readr(fhpara,1)';
closeall fhpara;

open fhretc = ^lretc for read;
_csretcode = readr(fhretc,1);
closeall fhretc;

/* _csretcode = 0;*/

if not(_csretcode);     /** calculate Hessian only if posterior mode is in shape **/

    /** print posterior mode **/
    if _verbose;

        "";
        "Parameter   | Estimate ";
        outmat = para_names'~para;
        let mask[1,2] = 0 1;
        let fmt[2,3]  =
                        "-*.*s " 10 6
                        "*.*lf " 10 4;
        d = printfm(outmat,mask,fmt);

        "";

        "Press Any Key to Continue"; k = keyw;
        " ";

    endif;      /** _verbose **/


    /********************************************************
    * Compute Hessian:
    * element by element, fine tune with dxscale
    ********************************************************/
    comphess:

    /** dxscale **/
    ndx      = 6;                   /** number of grid **/
    dx       = exp(-seqa(5,2,ndx)); /** grid **/
    dxscale  = ones(npara,1);       /** scale can be set differently across parameters **/

    /** set storage **/
    hessian  = zeros(npara,npara);
    hessdiag = zeros(ndx,1);

    /** evaluate posteior at its mode **/
    fx = fcn(para);

    if _verbose==1;
        "Computing Hessian..";
        "-------------------";
        "";
        "  diagonal elements:     ";;ftos(npara,"%*.*lf",6,0);
        "  ";;
    endif;

    /** diagonal elements **/
    seli = 1;
    do until seli > npara;

        /** print process **/
        if _verbose==1;
            ".";; pause(1e-20);
        elseif _verbose==2;
            "Hessian Element    (" seli seli ")";
            pause(1e-10);
        endif;  /** _verbose **/

        /** evaluation **/
        i=1;
        do until i > ndx;
            paradx = para;
            parady = para;
            paradx[seli] = paradx[seli] + dx[i]*dxscale[seli];  /** forward **/
            parady[seli] = parady[seli] - dx[i]*dxscale[seli];  /** backward **/
            fdx = fcn(paradx);
            fdy = fcn(parady);
            hessdiag[i] = -( 2*fx - fdx - fdy)/(dx[i]*dxscale[seli])^2;
            i = i+1;
        endo;

        /** negative of Hessian (average of values at 3rd and 4th grid) **/
        hessian[seli,seli] = -0.5*(hessdiag[3]+hessdiag[4]);

        /** print result **/
        if _verbose==2;
            "";
            "   Values     ";; -hessdiag';
            "   Value Used:";; hessian[seli,seli];
            "";
        endif;      /** _verbose **/

        seli = seli+1;
    endo;

    /** off-diagonal elements                                   **/
    /** Make sure that correlations are between -1 and 1        **/
    /** errorij contains the index of elements that are invalid **/
    if _verbose==1;
        "";
        "  off-diagonal elements: ";;ftos(npara*(npara-1)/2,"%*.*lf",6,0);
        "  ";;
    endif;

    errorij = {};

    seli = 1; vidx = 1;
    do until seli > npara;

        selj = seli+1;          /** upper triangular part **/

        do until selj > npara;

            /** print process **/
            if _verbose==1;
              ".";; pause(1e-20);
              if fmod(vidx,50)==0; "";"  ";; endif;
              vidx=vidx+1;
            elseif _verbose==2;
             "Hessian Element    (";; seli;; selj;; ")";
              pause(1e-10);
            endif;          /** _verbose **/

            /** evaluation **/
            i=1;
            do until i > ndx;
                paradx = para;
                parady = para;
                paradx[seli] = paradx[seli] + dx[i]*dxscale[seli];
                parady[selj] = parady[selj] - dx[i]*dxscale[selj];
                paradxdy = paradx;
                paradxdy[selj] = paradxdy[selj] - dx[i]*dxscale[selj];
                fdx = fcn(paradx);
                fdy = fcn(parady);
                fdxdy = fcn(paradxdy);
                hessdiag[i] = -( fx - fdx - fdy + fdxdy )/(dx[i]*dx[i]*dxscale[seli]*dxscale[selj]);
              i = i+1;
            endo;

            /** print result **/
            if _verbose==2;
                "";
                "  Values           ";; -hessdiag';
            endif;          /** _verbose **/

            /** negative of Hessian (average of values at 3rd and 4th grid) **/
            hessian[seli,selj] = -0.5*(hessdiag[3]+hessdiag[4]);

            /** calculate correlation **/
            if ( hessian[seli,seli] == 0 ) or (hessian[selj,selj] == 0);        /** zero variance **/
                corrij = 0;
            else;                   /** (covariance)/(standard deviations) **/
                corrij = hessian[seli,selj]/sqrt(hessian[seli,seli]*hessian[selj,selj]);
            endif;

            /** adjust correlation **/
            if abs(corrij)>0.98;                    /** close to 1 **/
                hessian[seli,selj]=0.9*sqrt(hessian[seli,seli]*hessian[selj,selj]);
                errorij = errorij |(seli~selj~corrij);
            elseif abs(corrij)<0.005;           /** close to 0 **/
                hessian[seli,selj]=0;
            endif;

            /** symmetrize **/
            hessian[selj,seli] = hessian[seli,selj];

            /** print result **/
            if _verbose==2;
                "  Value Used:      ";; hessian[seli,selj];
                "  Correlation:     ";; corrij;
                "  Number of Errors:";; rows(errorij);
                "";
            endif;  /** _verbose **/

            selj=selj+1;
        endo;
        seli = seli+1;
    endo;

    /** print errors in correlation **/
    if _verbose==2;
          "Errors";  errorij;
    endif;  /** _verbose **/

    hessian = real(hessian);

/*************************************************************
FIX hessian elements here
*************************************************************/
/*
hessian[.,1] = zeros(npara,1);
hessian[1,.] = zeros(1,npara);
hessian[1,1] = 1/pstdd[1]^2;

hessian[.,4] = zeros(npara,1);
hessian[4,.] = zeros(1,npara);
hessian[4,4] = 1/pstdd[4]^2;
*/
/*************************************************************
FIX hessian elements here
*************************************************************/
 
   /** Save output **/
    ohess = postpath $+ "\\MSleadershipfm_hess" $+ pmsel;
    create fhhess=^ohess with hess, npara, 8;
    wr = writer(fhhess,hessian);
    closeall fhhess;


    /********************************************************
    * Compute covariance matrix and penalty (posterior mode)
    ********************************************************/

    /** load (negative) Hessian **/
    lhess = postpath $+ "\\MSleadershipfm_hess" $+ pmsel;
    open fhhess = ^lhess for read;
    HHM = readr(fhhess,npara);
    closeall fhhess;

    /** rank of Hessian (theoretical) **/
    rankHHM = sumc(pmaskinv);
  
    /** inverse by Singular Value Decomposition (SVD) **/
    {u , s, v} = svd1(HHM);
    invHHMdet = 1;

    i = 1;
    do until i > npara;
       if i > rankHHM;
          s[i,i] = 0;
       else;
          s[i,i]    = 1/s[i,i];
          invHHMdet = invHHMdet*s[i,i];         /** determinant for nonsingular part **/
       endif;
       i = i+1;
    endo;

    invHHM  = u*s*u';            /** inverse of negative Hessian: covariance **/
    sigmult = u*sqrt(s);         /** square root of covariance matrix at posterior mode **/
    penalt  = (rankHHM/2)*ln(2*pi) + 0.5*ln(invHHMdet); /** posterior mode penalty **/

    /** print result **/
    "";"";
    "Determinant of variance matrix";
    invHHMdet;
    "";
    "s.e. at the posterior mode";
    sqrt(diag(invHHM));
    "";
    "Post Mode Penalty";
    penalt;
    "";

    /** Save sigmult (square root of posterior mode covariance) **/
    omult = postpath $+ "\\MSleadershipfm_mult" $+pmsel;
    create fhmult=^omult with MULT, npara, 8;
    wr = writer(fhmult,sigmult);
    closeall fhmult;

else;
    "Check posterior mode.";

endif;      /** _csretcode **/


/****************************************************
* PROCEDURES: FCN
* evaluates posterior density
****************************************************/

proc (1) = fcn(para);
local lnpY, lnprio, obserror, obsvar, retcode, modelpara,probsT,ssT_draw,states;

modelpara = para;
modelpara = modelpara.*pmaskinv + pfix.*pmask;


{lnpY,retcode,obserror,obsvar,probsT,ssT_draw} = Discretion_evallp(modelpara,YY,msel); 

/** Evaluate the Prior distribution **/
lnprio = priodens(modelpara,pmean,pstdd,pshape);

retp(real(lnpY+lnprio)); 

endp;
