/******************************************************************************
* NKMP_MH.G
* generates draws from the posterior distribution
* via the random walk Metropolis-Hastings algorithm
*
* upadted:  06/16/2005 
******************************************************************************/
new;
/** set seed **/
/** rndseed 1234; **/

/********************************************************
* House cleaning and set defaults
********************************************************/
library pgraph, user;
cls;
outwidth 256;

format /mat /on /mb1 /ros 16,8;

/** Global defaults **/
_verbose = 1;       /** 1 for verbose **/

/** only the following 2 lines are model dependent **/
#include 1MSleadershipfmmod.src;
#include 2MSleadershipfmlh.src;
#include 3MSleadershipfmpara.src;
#include spec.g;
#include dsgesel2.src;
#include pathspec.g;
#include matop.src;

/********************************************************
* Configure the Metropolis algorithm
********************************************************/
pmsel   = "";
mhrun   = "11";

nblocks = 5400;
nsim    = 100;

cc0     = 1;
cc      = 0.20;  

/* 
** use cc = 0.25 for s5t1
*/
appendmode = 1;

/** start log file **/
outstr = resupath $+ "\\MSleadershipfm_mh.log";
output file = ^outstr reset;

/********************************************************
* Import data series
********************************************************/
#include loaddata.g; 

nobs  = rows(YY);          /** number of observations **/

/********************************************************
*  Import prior specification
********************************************************/
priorfile = "MSleadershipfmused" $+ prioname $+ ".txt";
load path = ^priopath prior[] = ^priorfile;
prior = reshape(prior,rows(prior)/5,5);

{prior,para_names} = psel(prior,msel);

npara    = rows(prior);
pshape   = prior[.,1];
pmean    = prior[.,2];
pstdd    = prior[.,3];
pmask    = prior[.,4];
pfix     = prior[.,5];
pmaskinv = 1-pmask;
pshape   = pshape.*pmaskinv;
/** parameter specification: transformation and bounds **/
_trspec = prtr(msel);
_trspec[.,1] = _trspec[.,1].* pmaskinv;
_bounds      = _trspec[.,2 3];

/********************************************************
* load Posterior Mode Estimates
********************************************************/
lpara = postpath $+ "\\MSleadershipfm_pm" $+ pmsel;
lretc = postpath $+ "\\MSleadershipfm_pmret" $+ pmsel;

/** check posterior mode results **/
if not(exist(lpara $+ ".dat"));
    "";
    "Please run posterior maximization program first.";
    "";
    end;
endif;

/** open **/
open fhpara = ^lpara for read;
para = readr(fhpara,1);
closeall fhpara;

para  = para';          /** paraest was saved as row vector in _pm.g **/
npara = rows(para);

open fhretc = ^lretc for read;
_csretcode = readr(fhretc,1);
closeall fhretc;

/********************************************************
* load multiplication matrix for parameters
********************************************************/
lmult = postpath $+ "\\MSleadershipfm_mult" $+ pmsel ;       /** sigmult **/
lhess = postpath $+ "\\MSleadershipfm_hess" $+ pmsel;     /** HHM **/

/** check Hessian results **/
if not(exist(lmult $+ ".dat"));
    "";
    "Please run Hessian finding program first.";
    "";
    end;
endif;

/** open **/
open fhmult = ^lmult for read;
SIGSCALE    = readr( fhmult,npara );
closeall fhmult;

open fhhess = ^lhess for read;
SIGPROPINV  = readr( fhhess,npara );
closeall fhhess;

/** rank of Hessian (theoretical) **/
SIGPROPdim = sumc(pmaskinv);

/** inverse by Singular Value Decomposition (SVD) **/
{u , s, v} = svd1(SIGPROPINV);
SIGPROPlndet = 0;

/** log determinant for nonsingular part **/
i = 1;
do until i > npara;
   if i <= SIGPROPdim;
      s[i,i]    = 1/s[i,i];
      SIGPROPlndet = SIGPROPlndet + ln(s[i,i]);
   endif;
   i = i+1;
endo;

/********************************************************
* Initialize Output files
********************************************************/

opath   = parapath $+ "\\mhrun" $+ mhrun;
chkdir  = chk_dir(opath);

/** print specification **/
if _verbose;

    "";
    "===================================";
    "Model            " $lmodel;
    "Data set         " $dataselstr;
    "Prior            " $lprior;
    "Subsample        " $subTstr;
    "MHrun Index      " $mhrun;
    "Scaling of Hess  " cc;
    "";
    "===================================";
    "";

    "";
    "Parameter   | Estimate ";
    outmat = para_names'~para;
    let mask[1,2] = 0 1;
    let fmt[2,3]  =
                    "-*.*s " 10 6
                    "*.*lf " 10 4;
    d = printfm(outmat,mask,fmt);
    {lnpost0, lnpy0} = fcn2(para);

    "";

    "Press Key to Continue";
    k = keyw;
    "";

endif;      /** _verbose **/


statname = { POST LIKE PROP PROPPOST REJECT };

ostat = opath $+ "\\MSleadershipfm_st";
opara = opath $+ "\\MSleadershipfm_pa";

/** check if append is possible **/
if not(exist(opara $+ ".dat"));
    appendmode=0;
endif;

if appendmode ne 1;

  /*******************************
  ** create new output files
  *******************************/

  "-----------------------------------------------------------";
  "Creating new output files";
  "Press Key to Continue";
  "-----------------------------------------------------------";
  k = keyw;
  "";

  create fhstat = ^ostat with ^statname, 5, 4;
  create fhpara = ^opara with ^para_names, npara, 4;

else;

  /*******************************
  ** open existing output files
  *******************************/

  "-----------------------------------------------------------";
  "Opening Existing Output Files";
  "Press Key to Continue";
  "-----------------------------------------------------------";

  open fhpara = ^opara for update;
  open fhstat = ^ostat for update;

  rowind = seekr(fhpara,rowsf(fhpara));
  rowind = seekr(fhstat,rowsf(fhpara));

  para    = readr(fhpara,1)';
  statsim = readr(fhstat,1);

  "Number of draws in File" rowind;
  "";

  rowind = seekr(fhpara,0);
  rowind = seekr(fhstat,0);

  cc0 = 0;

  "Parameter   | Last Draw ";
  outmat = para_names'~para;
  let mask[1,2] = 0 1;
  let fmt[2,3] =
               "-*.*s " 10 6
               "*.*lf " 10 4;
  d = printfm(outmat,mask,fmt);

  "";
  "Press Key to Continue";
  "-----------------------------------------------------------";
  k = keyw;
  
endif;

/********************************************************
* Metropolis-Hastings algorithm
********************************************************/

/********************************************************
** try overdispersed starting values for kappa and psi1 (2,3)
**
** 1) tau = 3, psi2 = 0.1     pm2, mhrun3
** 2) tau = 3, psi2 = 2       pm1, mhrun4
** 3) tau = 0.4, psi2 = 0.1   pm2, mhrun5
** 4) tau = 0.62, psi2 = 2    pm1, mhrun6 
*/

/*
cc0     = 0;
para[1] = 0.62;
para[4] = 2;
*/

/** initialize PARANEW **/
/** trembles the posterior mode, and posterior is large enough **/
valid = 0;
do until valid;
   paranew = para + cc0*(sigscale*rndn(npara,1));
   paranew = paranew.*pmaskinv + pfix.*pmask;

   {postnew, likenew} = fcn2(paranew);
   propdens  = -0.5*SIGPROPDIM*ln(2*pi) - 0.5*SIGPROPlndet - 0.5*SIGPROPDIM*ln(cc0^2)
               -0.5*(paranew - para)'*SIGPROPINV*(paranew - para)/cc0^2;
   if postnew > -10000;
      valid = 1;
   endif;
endo;

/** block loop **/
iblock = 1;
do until iblock > nblocks;
    
   if _verbose;
   "";
   "----------------------------------------";
   "Block" iblock "of" nblocks;
   pause(1e-20);
   endif;

   /** set start time **/
   tstart = date;

   /** storages **/
   parasim = zeros(nsim,npara);
   likesim = zeros(nsim,1);
   postsim = zeros(nsim,1);
   rej     = zeros(nsim,1);
   propsim = zeros(nsim,1);
   proppostsim = zeros(nsim,1);

   if (iblock == 1) and (appendmode ne 1);
      /** save initial value as 1st draw in 1st block **/
      j = 2;
      likesim[1,1] = likenew;
      postsim[1,1] = postnew;
      propsim[1,1] = propdens;
      proppostsim[1,1] = postnew;
      parasim[1,.] = paranew';
      postold      = postnew;
      likeold      = likenew;
      paraold      = paranew;
   elseif (iblock == 1) and (appendmode == 1);
      j = 1;
      postold      = postnew;
      likeold      = likenew;
      paraold      = paranew;
   else;
      /** don't save initial values **/
      j=1;
   endif;

   /** with-in block loop **/
    do until j > nsim;

      /** propose **/
      paranew = paraold + cc*(sigscale*rndn(npara,1));
      paranew = paranew.*pmaskinv + pfix.*pmask;

      /** evaluate at the candidate **/
      {postnew, likenew} = fcn2(paranew);
      propdens = -0.5*SIGPROPDIM*ln(2*pi) - 0.5*SIGPROPlndet - 0.5*SIGPROPDIM*ln(cc^2)
                 -0.5*(paranew - paraold)'*SIGPROPINV*(paranew - paraold)/cc^2;
      propsim[j,1] = propdens;
      proppostsim[j,1] = postnew;

      /** accept/reject rule  **/
      r = minc(1 | exp( postnew - postold));

      if rndu(1,1) < r;

          /** accept proposed jump **/
          postsim[j,1] = postnew;
          likesim[j,1] = likenew;
          parasim[j,.] = paranew';

          /** updates **/
          paraold = paranew;
          postold = postnew;
          likeold = likenew;

      else;

          /** reject proposed jump **/
          likesim[j,1] = likeold;
          postsim[j,1] = postold;
          parasim[j,.] = paraold';
          rej[j] = 1;

      endif;

      /**
      "Block" iblock "of" nblocks;
      "Simulation step" j "of" nsim;
      "Proposal Dens  " propsim[j,1];
      "Parameters     "; parasim[j,.];
      **/

      /*  if _verbose;
            if fmod(j,100)==0;".";; pause(1e-20); endif;
        endif;*/

      j = j+1;
    endo;               /** with-in block  **/

   if _verbose;
       " ";
       "Rejection: perct "  meanc(rej);
       "Likelihood       "  meanc(likesim);
       "Posterior        "  meanc(postsim);
       "Parameters       "; meanc(parasim)';
       "";
       pause(1e-20);
   endif;  /** verbose **/

   /** set finish time **/
   tfin = date;

   if _verbose;
       "Elapsed Time"; etstr(ethsec(tstart,tfin));
       "";
   endif;  /** verbose **/

   /** save output **/
   wr = writer(fhstat,postsim~likesim~propsim~proppostsim~rej);
   wr = writer(fhpara,parasim);

   iblock = iblock+1;

endo;           /** block **/

closeall fhstat, fhpara;


//#include sw_mhdd.g;
/** close log file **/
output file = ^outstr off;


/****************************************************
* PROCEDURES: FCN2
* evaluates posterior density given parameter bounds
****************************************************/
proc (2) = fcn2(para);
local lnpY, lnprio, obserror, obsvar, lnpost, retcode, parabd_ind1, parabd_ind2,parabd_ind3,probsT,ssT_draw,modelpara,states;

    parabd_ind1 = para .> _bounds[.,1];         /** lower bounds **/
    parabd_ind2 = para .< _bounds[.,2];         /** upper bounds **/
    parabd_ind3 = (para[32]+para[33]) .<=1;  
    parabd_ind1 = delif(parabd_ind1,pmask);
    parabd_ind2 = delif(parabd_ind2,pmask);
      

    if (parabd_ind1 == 1) and (parabd_ind2 == 1) and (parabd_ind3 == 1);   /** in bounds **/

       {lnpY,retcode,obserror,obsvar,probsT,ssT_draw} = Discretion_evallp(para,YY,msel); 
        
  
        lnpy    = real(lnpy);
        lnprio  = priodens( para, pmean, pstdd, pshape);
        lnpost  = lnpy + real(lnprio);

    else;                                                           /** out of bounds **/

        lnpost = -1E6;
        lnpy   = -1E20;

    endif;

retp(lnpost,lnpy);
endp;
