/* filename:    sw_mhdd.g
** description: Computes numerical approximations to the data density
**              based on MCMC output
*/

/******************************************************************
**         Load Parameter Draws from MH Output
*/
new;
closeall;
library pgraph, user;
cls;
/** Global defaults **/
_verbose = 1;       /** 1 for verbose **/
pmsel   = "";
mhrun   = "11";

#include 1MSleadershipfmmod.src;
#include 2MSleadershipfmlh.src;
#include 3MSleadershipfmpara.src;
#include spec.g;
#include dsgesel2.src;
#include pathspec.g;
#include matop.src;

/********************************************************
*  Import prior specification
********************************************************/
priorfile = "MSleadershipfmused" $+ prioname $+ ".txt";
load path = ^priopath prior[] = ^priorfile;
prior = reshape(prior,rows(prior)/5,5);

{prior,para_names} = psel(prior,msel);

pshape   = prior[.,1];
pmean    = prior[.,2];
pstdd    = prior[.,3];
pmask    = prior[.,4];
pfix     = prior[.,5];
pmaskinv = 1-pmask;
pshape   = pshape.*pmaskinv;

/** parameter specification: transformation and bounds **/
_trspec      = prtr(msel);
_trspec[.,1] = _trspec[.,1].*pmaskinv;
_bounds      = _trspec[.,2 3];

/** set environments **/
nburn   = 1;   /** initial draws for burning**/
nblock  = 1200;    /** read block size **/
hmax    = 2;
//densfac = -1000;    //densfac = meanc(postsim);
npara1  = sumc(pmaskinv);

/******************************************************************
* Initialize Output files
*
* output file contains
*   {npara1~parahat'~lowtrvec(parasig)'~pym3'~suminvlikeerror}
*    1      npara1   npara*(npara+1)/2  9     9
******************************************************************/
opath   = parapath $+ "\\mhrun" $+ mhrun;
chkdir  = chk_dir(opath);

noutdim = 1 + npara1 + npara1*(npara1+1)/2 + 9 + 9;
omhdd   = opath $+ "\\MSleadershipfm_mhdd";
create fhmhdd=^omhdd with mhdd, noutdim, 8;

/******************************************************************
* Load Parameter Draws from MH Output
******************************************************************/
lpath  = parapath $+ "\\mhrun" $+ mhrun;
chkdir = chk_dir(opath);
lstat  = lpath $+  "\\MSleadershipfm_st";
lpara  = lpath $+  "\\MSleadershipfm_pa";

open fhpara = ^lpara for read;
open fhstat = ^lstat for read;

postsim = readr( fhstat, nblock );
densfac = meanc(postsim[.,1]); 

/******************************************************************
* Method:  modified harmonic mean
******************************************************************/

/** decile of chi2(npara1) **/
p = seqa(0.1,0.1,9); 
pcrit = cdfchii(p,npara1);

/*******************************************************
* Part 1: Compute posterior moments for parameters
*******************************************************/

eofloop = 0;
ndraws  = 0;
sumpara  = zeros(npara1,1);
sumpara2 = zeros(npara1,npara1);

/** discard initial draws **/
pararow = seekr(fhpara,nburn);
statrow = seekr(fhstat,nburn);



do until eofloop; 

   /** read block from MH output **/
 
   parasim = readr( fhpara, nblock );
   parasim = delif(parasim',pmask)';

   sumpara  = sumpara + sumc(parasim);      /** sum **/
   sumpara2 = sumpara2 + parasim'*parasim;  /** squared (cross product) sum **/
   ndraws   = ndraws + rows(parasim);       /** number of draws **/
   


   eofloop = eof(fhpara);

endo;


/** posterior moments **/
parahat       = sumpara / ndraws;                           /** mean **/
parasig       = sumpara2 / ndraws - parahat*parahat';       /** variance **/

/** check diagonal elements **/
i = 1;
do until i > rows(parasig);
   if parasig[i,i] < 1E-6;
      parasig[i,i] = 1E-6;
   endif;
   i = i+1;
endo;

parasiginv   = inv(parasig);
parasiglndet = ln(det(parasig));
ltrvecsig = lowtrvec(parasig);

/*******************************************************
* Part 2:  Compute Density estimate
*******************************************************/

/** initialize **/
eofloop = 0;
ndraws  = 0;
suminvlike = zeros(rows(p),1);
laginvlike = zeros(hmax,rows(p));
gaminvlike = zeros(hmax,rows(p));

/** discard initial draws **/
pararow = seekr(fhpara,nburn);
statrow = seekr(fhstat,nburn);



do until eofloop;

    /** read block from MH output **/
   statsim = readr( fhstat, nblock);
   parasim = readr( fhpara, nblock);

   /** get rid of fixed parameters **/
   parasim = delif(parasim',pmask)';

   i = 1;
   do until i > rows(parasim);
    
      lnfpara = -ln(p) - 0.5*npara1*ln(2*pi) - 0.5*parasiglndet
                 -0.5*(parasim[i,.] - parahat')*parasiginv*(parasim[i,.]' - parahat);
      indpara = ((parasim[i,.] - parahat')*parasiginv*(parasim[i,.]' - parahat) .< pcrit );
      invlike = exp(lnfpara - statsim[i,1] + densfac).*indpara;
   
      laginvlike = invlike' | laginvlike[1:hmax-1,.];
      
 
      gaminvlike = gaminvlike + laginvlike.*invlike';
      suminvlike = suminvlike + invlike;
      i = i+1;
   endo;

    ndraws = ndraws + rows(parasim);
    eofloop = eof(fhpara);

endo;

meaninvlike = suminvlike/ndraws;  
gaminvlike  = gaminvlike/ndraws - (meaninvlike^2)';

/** Compute Standard Errors **/
suminvlikeerror = gaminvlike[1,.];


j = 2;
do until j > hmax;
   suminvlikeerror = suminvlikeerror + 2*gaminvlike[j,.]*(1-(j-1)/hmax);
   j = j+1;
endo;
suminvlikeerror = 100*sqrt(suminvlikeerror/ndraws)./meaninvlike' ;

pym3    = densfac - ln(meaninvlike); 

if _verbose;
    "";"";
    "Model:  ";; $lmodel;
    "Prior:  ";; $lprior;
    "MhRun:  ";; $mhrun;
    "Dataset:";; $dataselstr;
    "======================";
    "";

    "Marginal Data Density";; pym3;
    "Average Across p     ";; meanc(pym3);
    "Simulation Error [Percent]";; suminvlikeerror';
endif;

wr = writer(fhmhdd,npara1~parahat'~lowtrvec(parasig)'~pym3'~suminvlikeerror);

closeall fhpara, fhstat, fhmhdd;
